!
! Levels.
!

module dims_levels

  use const_ec_v, only : lme, a_ec, b_ec
  
  use dims_grid, only : nregions
  
  implicit none
  
  ! --- in/out ------------------------------
  
  public
  
  
  ! --- const -------------------------------
  
  
  integer, parameter ::  lm(nregions) = 34

  ! Only lowest layers are used for convec calculations (sub files)
  ! to avoid memory problems.
  ! Quote Michiel van Weele:
  !     "..., convectie kan wel tot ~70 hPa (19km) gaan in de tropen.
  !     Ik denk dat je de bovenste 33 lagen veilig kan vernullen."
  ! Thus, the lowest 91 - 33 = 58  in ml91 should be enough.
  ! In tropo34, this is the lowest 23 layers.
  integer,parameter :: lmax_conv = 23

  ! select ECMWF half levels;
  ! TM levels are number bottom-up:
  integer, parameter   ::  echlevs(0:lm(1)) = (/ &
                                                             91,     &
                             89,         86,         83,         80, &
                                     77,         74,         71,     &
                                 68,         65,         62,         &
                             59,         56,     54,     52,     50, &
                                 48,     46,     44,     42,     40, &
                                 38,     36,     34,     32,         &
                             29,         26,         23,         20, &
                                     17,         14,         11,     &
                                  8,          5,                  0 /)

  real, parameter  :: bt(1:lm(1)+1) = b_ec(echlevs)
  real, parameter  :: at(1:lm(1)+1) = a_ec(echlevs)


end module dims_levels

