!#################################################################
!
! TM5 save files
!
!### macro's #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#define slopes
!
!#################################################################


module save_files

  use GO, only : gol, goErr, goPr
  
  implicit none
  
  
  ! --- in/out ----------------------------------
  
  private
  
  public  ::  ReadTracer_saved, ReadCO2_saved
  
  
  ! --- const --------------------------------
    
  character(len=*), parameter  ::  mname = 'savefiles'

  

contains


  ! =====================================================================

  subroutine ReadCO2_saved( region, fname, status)

    use file_hdf   , only : THdfFile, TSds, Init, Done, ReadAttribute, ReadData
    use grid       , only : TllGridInfo, TLevelInfo, Init, Done, Fill3D
    use dims       , only : im, jm, lm
    use chem_param , only : ntrace, ntracet, fscale, names, ntrace_chem
    use partools   , only : myid, root
    use partools   , only : previous_par, tracer_active, ntracetloc, tracer_loc
    use partools   , only : Par_Barrier
#ifdef MPI
    use mpi_comm   , only : scatter_after_read_k_3d
#endif
    use meteodata  , only : lli, levi, sp_dat
    use tracer_data, only : mass_dat, chem_dat

    ! --- in/out -------------------------------
    
    integer, intent(in)           ::  region
    character(len=*), intent(in)  ::  fname
    integer, intent(out)          ::  status
    
    ! --- const -------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/ReadCO2_saved'
    
    ! --- local -------------------------------
    
    integer                             ::  imr, jmr, lmr

    real, dimension(:,:,:), pointer     ::  sp
    real, dimension(:,:,:), pointer     ::  m
    real, dimension(:,:,:,:), pointer   ::  rm

    type(THdfFile)                   ::  hdf
    type(TSds)                       ::  sds_rm, sds_m
    integer                          ::  n, n_loc
    integer                          ::  ntrace_in, ntracet_in
    integer                          ::  n_in, nl
    character(len=8), allocatable    ::  names_in(:)
    integer                          ::  xbeg_in, ybeg_in
    real                             ::  dx_in, dy_in
    integer                          ::  im_in, jm_in, lm_in
    real, allocatable                ::  at_in(:), bt_in(:)
    type(TllGridInfo)                ::  lli_in
    type(TLevelInfo)                 ::  levi_in
    real, allocatable                ::  rm_in(:,:,:,:)
    real, allocatable                ::  msave(:,:,:)
    integer                          ::  l
    logical                          ::  with_tm4

    ! --- begin -------------------------------
    
    ! should be parallel over tracers ...
    if ( previous_par(region) /= 'tracer' ) then
      write (gol,'("expected parallel over tracers:")'); call goErr
      write (gol,'("  region       : ",i3)') region; call goErr
      write (gol,'("  previous_par : ",a)') trim(previous_par(region)); call goErr
      write (gol,'("in ",a)') rname; call goErr; status=1; return
    end if
    
    ! grid sizes
    imr = im(region)
    jmr = jm(region)
    lmr = lm(region)
    
    ! set pointers
    sp => sp_dat(region)%data
    m  => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t

    ! open hdf file (all pe)
    call Init( hdf, fname, 'read', status )
    IF_NOTOK_RETURN(status=1)

    ! read list with tracer names in file:
    call ReadAttribute( hdf, 'ntrace', ntrace_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'ntracet', ntracet_in, status )
    IF_NOTOK_RETURN(status=1)
    allocate( names_in(ntrace_in) )
    call ReadAttribute( hdf, 'names', names_in, status )
    IF_NOTOK_RETURN(status=1)

    ! temporary storage:
    allocate( msave(imr,jmr,lmr) )
    allocate( rm_in(imr,jmr,lmr,1) )

    ! open air masses:
    call Init( sds_m, hdf, 'm', status )  
    IF_NOTOK_RETURN(status=1)
    
    ! open tracer masses:
    call Init( sds_rm, hdf, 'rm', status )  
    IF_NOTOK_RETURN(status=1)

    ! loop over all tracers:
    do n = 1, ntrace
    
      ! search input record:
      n_in = -1
      do nl = 1, ntrace_in
        if ( names(n) == names_in(nl) ) then
          n_in = nl
          exit
        end if
      end do
      ! not found ? 
      if ( n_in < 0 ) then
        write (gol,'("WARNING - tracer not found in save file:")'); call goPr
        write (gol,'("WARNING -   file     : ",a)') trim(fname); call goPr
        write (gol,'("WARNING -   tracer   : ",i3," ",a)') n, trim(names(n)); call goPr
        cycle
      end if

      ! not on this proces ?
      if ( .not. tracer_active(n) ) cycle

      ! read 3D mass field: 
      call ReadData( sds_m, msave, status)  ! kg
      IF_NOTOK_RETURN(status=1)

      ! read 3D tracer field: 
      call ReadData( sds_rm, rm_in, status, start=(/0,0,0,n_in-1/) )  ! kg
      IF_NOTOK_RETURN(status=1)
      
      ! local index:
      n_loc = tracer_loc(n)

      ! convert grids and levels:
      rm(1:imr,1:jmr,1:lmr,n_loc)= rm_in(:,:,:,1)/msave(1:imr,1:jmr,1:lmr)*m(1:imr,1:jmr,1:lmr)

      ! info
      write (gol,*) 'initialised tracer ', n, names(n), '; max mass, max airmaass frac, max vmr : ', &
                         maxval(rm(1:imr,1:jmr,1:lmr,n_loc)), &
                         maxval(m(1:imr,1:jmr,1:lmr)/msave(1:imr,1:jmr,1:lmr)),&
                         maxval(rm(1:imr,1:jmr,1:lmr,n_loc)/m(1:imr,1:jmr,1:lmr)*fscale(n)); call goPr
         
    end do
    
    ! clear
    deallocate( names_in )
    deallocate( rm_in )
    deallocate( msave )
    
    ! close data sets:
    call Done( sds_m, status )  
    IF_NOTOK_RETURN(status=1)
    call Done( sds_rm, status )  
    IF_NOTOK_RETURN(status=1)
    
    ! close hdf file (all pe)
    call Done( hdf, status )
    IF_NOTOK_RETURN(status=1)
    
    ! wait ...
    call Par_Barrier
    
    ! ok
    status = 0
    
  end subroutine ReadCO2_saved

  subroutine ReadTracer_saved( region, fname, status, tm4 )

    use file_hdf   , only : THdfFile, TSds, Init, Done, ReadAttribute, ReadData
    use grid       , only : TllGridInfo, TLevelInfo, Init, Done, Fill3D
    use dims       , only : im, jm, lm
    use chem_param , only : ntrace, ntracet, fscale, names, ntrace_chem
    use partools   , only : myid, root
    use partools   , only : previous_par, tracer_active, ntracetloc, tracer_loc
    use partools   , only : Par_Barrier
#ifdef MPI
    use mpi_comm   , only : scatter_after_read_k_3d
#endif
    use meteodata  , only : lli, levi, sp_dat
    use tracer_data, only : mass_dat, chem_dat

    ! --- in/out -------------------------------
    
    integer, intent(in)           ::  region
    character(len=*), intent(in)  ::  fname
    integer, intent(out)          ::  status
    
    logical, intent(in), optional     ::  tm4
    
    ! --- const -------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/ReadTracer_saved'
    
    ! 31 (from 60) layer defintition for TM4 save files:
    real, parameter  ::  at_tm4_60_31(0:31) = (/ &
      0.,             7.367743,       467.333588,     1385.912598,    2887.696533,    &
      4941.778320,    6144.314941,    8802.356445,    10209.500977,   11632.758789,   &
      14411.124023,   15706.447266,   16899.468750,   17961.357422,   18864.750000,   &
      19584.330078,   20097.402344,   20384.480469,   20222.205078,   19755.109375,   &
      19027.695313,   18045.183594,   16819.474609,   15379.805664,   13775.325195,   &
      12077.446289,   10376.126953,   8765.053711,    6018.019531,    3960.291504,    &
      2082.273926,    0./)
    real, parameter  ::  bt_tm4_60_31(0:31) = (/ &
      1.00000000,     0.99401945,     0.96764523,     0.93194032,     0.87965691,     &
      0.81125343,     0.77159661,     0.68326861,     0.63554746,     0.58616841,     &
      0.48477158,     0.43396294,     0.38389215,     0.33515489,     0.28832296,     &
      0.24393314,     0.20247590,     0.16438432,     0.09967469,     0.07353383,     &
      0.05169041,     0.03412116,     0.02067788,     0.01114291,     0.00508112,     &
      0.00181516,     0.00046139,     0.00007582,     0.00000000,     0.00000000,     &
      0.00000000,     0.00000000/)

    ! 34 (from 91) layer defintition for TM4 save files:
    real, parameter :: at_tm4_91_34(0:34) = (/ &
     0.000000,     6.575628,    336.772369,   1297.656128,   3010.146973,  &
     5422.802734,  8356.252930, 11543.166992, 14665.645508, 17385.595703, &
     19348.775391, 20319.011719, 20348.916016, 19919.796875, 19184.544922, &
     18191.029297, 16990.623047, 15638.053711, 14192.009766, 12713.897461, &
     11262.484375,  9873.560547,  8564.624023,  7341.469727,  6199.839355,  &
     4663.776367,  3358.425781,  2292.155518,  1463.163940,   857.945801,   &
     450.685791,   204.637451,    76.167656,    21.413612,     0.000000/)

    real, parameter  :: bt_tm4_91_34(0:34) = (/     &
     1.000000,     0.994204,     0.973466,     0.935157,     0.875518,     &
     0.795385,     0.698224,     0.589317,     0.475016,     0.362203,     &
     0.259554,     0.176091,     0.112979,     0.080777,     0.055474,     &
     0.036227,     0.022189,     0.012508,     0.006322,     0.002765,     &
     0.001000,     0.000279,     0.000055,     0.000000,     0.000000,     &
     0.000000,     0.000000,     0.000000,     0.000000,     0.000000,     &
     0.000000,     0.000000,     0.000000,     0.000000,     0.000000/)

    ! --- local -------------------------------
    
    integer                             ::  imr, jmr, lmr

    real, dimension(:,:,:), pointer     ::  sp
    real, dimension(:,:,:), pointer     ::  m
    real, dimension(:,:,:,:), pointer   ::  rm

    type(THdfFile)                   ::  hdf
    type(TSds)                       ::  sds_rm, sds_rmc
    integer                          ::  n, n_loc
    integer                          ::  ntrace_in, ntracet_in
    integer                          ::  n_in, nl
    character(len=8), allocatable    ::  names_in(:)
    integer                          ::  xbeg_in, ybeg_in
    real                             ::  dx_in, dy_in
    integer                          ::  im_in, jm_in, lm_in
    real, allocatable                ::  at_in(:), bt_in(:)
    type(TllGridInfo)                ::  lli_in
    type(TLevelInfo)                 ::  levi_in
    real, allocatable                ::  rm_in(:,:,:,:)
    real, allocatable                ::  rmt(:,:,:)
    integer                          ::  l
    logical                          ::  with_tm4
    character(len=40)                :: xx

    ! --- begin -------------------------------
    
    ! should be parallel over tracers ...
    if ( previous_par(region) /= 'tracer' ) then
      write (gol,'("expected parallel over tracers:")'); call goErr
      write (gol,'("  region       : ",i3)') region; call goErr
      write (gol,'("  previous_par : ",a)') trim(previous_par(region)); call goErr
      write (gol,'("in ",a)') rname; call goErr; status=1; return
    end if
    
    ! grid sizes
    imr = im(region)
    jmr = jm(region)
    lmr = lm(region)
    
    ! set pointers
    sp => sp_dat(region)%data
    m  => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t

    ! special input ?
    with_tm4 = .false.
    if ( present(tm4) ) with_tm4 = tm4

    ! open hdf file (all pe)
    write(xx,*) myid
    xx=adjustl(xx)
    call Init( hdf, trim(fname)//'_pe'//trim(xx), 'read', status )
    IF_NOTOK_RETURN(status=1)

    ! read list with tracer names in file:
    call ReadAttribute( hdf, 'ntrace', ntrace_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'ntracet', ntracet_in, status )
    IF_NOTOK_RETURN(status=1)
    IF (ntracet_in.ne.ntracet) THEN
      write (gol,'("ntracet does not correspond to ntracet in save file:")'); call goErr
      write (gol,'("  ntracet      : ",i3)') ntracet; call goErr
      write (gol,'("  ntracet_in   : ",i3)') ntracet_in; call goErr
      write (gol,'("Routine requires them to be equal")'); call goErr
      write (gol,'("in ",a)') rname; call goErr; status=1; return
    ELSE
      write (gol,'("Checked ntracet in save file: okay")'); call goPr
      write (gol,'("Assuming npes is equal for both runs!")') ; call goPr
    ENDIF

    ! read level definition:
    call ReadAttribute( hdf, 'lm', lm_in, status )
    IF_NOTOK_RETURN(status=1)
    allocate( at_in(0:lm_in) )
    allocate( bt_in(0:lm_in) )

    call ReadAttribute( hdf, 'at', at_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'bt', bt_in, status )
    IF_NOTOK_RETURN(status=1)

    call Init( levi_in, lm_in, at_in, bt_in, status )
    IF_NOTOK_RETURN(status=1)
    deallocate( at_in )
    deallocate( bt_in )   
    
    ! read grid definition:
    call ReadAttribute( hdf, 'im', im_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'jm', jm_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'xbeg', xbeg_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'dx', dx_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'ybeg', ybeg_in, status )
    IF_NOTOK_RETURN(status=1)
    call ReadAttribute( hdf, 'dy', dy_in, status )
    IF_NOTOK_RETURN(status=1)

    call Init( lli_in, xbeg_in+0.5*dx_in, dx_in, im_in, &
                       ybeg_in+0.5*dy_in, dy_in, jm_in, status )
    IF_NOTOK_RETURN(status=1)
    
    ! temporary storage:
    allocate( rm_in(im_in,jm_in,lm_in,1:ntracetloc) )
    
    ! open tracer masses:
    call Init( sds_rm, hdf, 'rm', status )  
    IF_NOTOK_RETURN(status=1)
    
    ! read 4D tracer field: 
    call ReadData( sds_rm, rm_in, status )  ! kg
    IF_NOTOK_RETURN(status=1)
    
    do n=1,ntracetloc    
        ! convert grids and levels:
        call Fill3D( lli(region), levi, 'n', sp(1:imr,1:jmr,1), rm(1:imr,1:jmr,1:lmr,n), &
                     lli_in, levi_in, rm_in(:,:,:,n), 'sum', status )
        IF_NOTOK_RETURN(status=1)

        ! info
        write (gol,*) 'initialised tracer ', n, names(n), '; max mass, max vmr : ', &
                           maxval(rm(1:imr,1:jmr,1:lmr,n)), &
                           maxval(rm(1:imr,1:jmr,1:lmr,n)/m(1:imr,1:jmr,1:lmr)*fscale(n)); call goPr
    end do     
    
    ! clear
    deallocate( rm_in )
    
    ! done with grid definitions:
    call Done( lli_in, status ) 
    IF_NOTOK_RETURN(status=1)
    call Done( levi_in, status ) 
    IF_NOTOK_RETURN(status=1)
    
    ! close data sets:
    call Done( sds_rm, status )  
    IF_NOTOK_RETURN(status=1)
    if ( .not. with_tm4 ) then
    end if
    
    ! close hdf file (all pe)
    call Done( hdf, status )
    IF_NOTOK_RETURN(status=1)
    
    ! wait ...
    call Par_Barrier
    
    ! ok
    status = 0
    
  end subroutine ReadTracer_saved
    



end module save_files
