!
! Input/output of meteofiles : prism version.
!

module tmm_mf_prism

  use GO, only : TDate
  
  implicit none
  
  ! --- in/out ----------------------------
  
  private
  
  public  ::  PRISM_Init, PRISM_Done

  public  ::  TMeteoFile_prism
  public  ::  Init, Done
  public  ::  ReadRecord

  public  ::  echam_tmin, echam_tmid, echam_tplus
  public  ::  echam_T, echam_shn, echam_lm
  public  ::  echam_lsp_tmin, echam_lsp_tmid, echam_lsp_tplus
  public  ::  echam_vor_tmid, echam_div_tmid
  public  ::  echam_temper_tmid

  
  ! --- const ------------------------------
  
  character(len=*), parameter  ::  mname = 'module tmm_mf_prism'
  
  !--- type ---------------------------------
  
  type TMeteoFile_prism
    integer                ::  id  ! dummy
  end type TMeteoFile_prism
  
  
  ! --- interfaces -------------------------
  
  interface Init
    module procedure mf_Init
  end interface

  interface Done
    module procedure mf_Done
  end interface

  interface ReadRecord
    module procedure mf_ReadRecord
  end interface


  ! --- local -----------------------------------
  
  ! echam spectral fields  
  integer  ::  echam_T  != 21
  integer  ::  echam_shn != (echam_T+1)*(echam_T+2)/2
  
  ! echam levels
  integer   ::  echam_lm != 19

  type(TDate)  ::  echam_tmin
  type(TDate)  ::  echam_tmid
  type(TDate)  ::  echam_tplus
  
!  real, allocatable    ::  echam_lsp_tmin (:,:)   ! (2,echam_shn)
!  real, allocatable    ::  echam_lsp_tmid (:,:)   ! (2,echam_shn)
!  real, allocatable    ::  echam_lsp_tplus(:,:)   ! (2,echam_shn)
!  real, allocatable    ::  echam_vor_tmid (:,:,:) ! (2,echam_shn,echam_lm)
!  real, allocatable    ::  echam_div_tmid (:,:,:) ! (2,echam_shn,echam_lm)

  ! >>> ahdoc fix >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  integer, parameter  ::  echam_T_FIXED  = 31
  integer, parameter  ::  echam_shn_FIXED = (echam_T_FIXED+1)*(echam_T_FIXED+2)/2
  integer, parameter   ::  echam_lm_FIXED = 39
  real    ::  echam_lsp_tmin (2,echam_shn_FIXED)
  real    ::  echam_lsp_tmid (2,echam_shn_FIXED)
  real    ::  echam_lsp_tplus(2,echam_shn_FIXED)
  real    ::  echam_vor_tmid (2,echam_shn_FIXED,echam_lm_FIXED)
  real    ::  echam_div_tmid (2,echam_shn_FIXED,echam_lm_FIXED)
  real    ::  echam_temper_tmid (2,echam_shn_FIXED,echam_lm_FIXED)
  ! <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

  ! *** levels *******************************************
    
  ! *** 19 ***
    
  real, parameter  :: echam_lev19_a(0:19) = (/   &
             0.000   ,  2000.000   ,  4000.000   ,  6046.111   , &
          8267.928   , 10609.510   , 12851.100   , 14698.500   , &
         15861.130   , 16116.240   , 15356.920   , 13621.460   , &
         11101.560   ,  8127.144   ,  5125.142   ,  2549.969   , &
           783.195   ,    0.0000   ,     0.000   ,     0.000        /)
    
  real, parameter  :: echam_lev19_b(0:19) = (/   &
             0.000000000, 0.00000000, 0.00000000, 0.0003389933, &
             0.003357187, 0.01307004, 0.03407715, 0.0706498300, &
             0.125916700, 0.20119540, 0.29551960, 0.4054092000, &
             0.524932200, 0.64610790, 0.75969840, 0.8564376000, &
             0.928746900, 0.97298520, 0.99228150, 1.0000000000    /)

  ! *** 39 ***
    
  real, parameter ::  echam_lev39_a(0:39) = (/ &
             0.000   ,     1.989182,     5.09726 ,    10.51217 , &
            18.75759 ,    30.39117 ,    47.1266  ,    71.56693 , &
           106.5414  ,   155.6308  ,   223.2727  ,   315.8983  , &
           442.4553  ,   613.6625  ,   843.0477  ,  1147.513   , &
          1547.971   ,  2070.053   ,  2744.883   ,  3609.916   , &
          4709.828   ,  6097.457   ,  7834.766   ,  9959.486   , &
         12317.67    , 14597.68    , 16441.67    , 17538.87    , &
         17708.88    , 16970.06    , 15587.9     , 13621.46    , &
         11101.56    ,  8127.155   ,  5125.142   ,  2549.969   , &
           783.195   ,     0.0     ,     0.0     ,     0.0          /)

  real, parameter  :: echam_lev39_b(0:39) = (/   &
             0.000000000, 0.000000000, 0.000000000, 0.000000000, &
             0.000000000, 0.000000000, 0.000000000, 0.000000000, &
             0.000000000, 0.000000000, 0.000000000, 0.000000000, &
             0.000000000, 0.000000000, 0.000000000, 0.000000000, &
             0.000000000, 0.000000000, 0.000000000, 0.000000000, &
             0.000000000, 0.000000000, 0.000000000, 0.0003389933, &
             0.003357187, 0.01307004 , 0.03407715 , 0.07064983  , &
             0.1259167  , 0.2011954  , 0.2955196  , 0.4054092   , &
             0.5249322  , 0.646108   , 0.7596984  , 0.8564376   , &
             0.9287469  , 0.9729852  , 0.9922815  , 1.0000000         /)
    


contains


  ! =============================================================
  
  !
  ! Init prism stuff.
  !
  ! NOTE: in the main program, some varibles should be set:
  !
  !   use tmm_mf_prism, only : echam_lm, echam_T
  !
  !   echam_lm = 39
  !   echam_T  = 31
  !
  
  subroutine PRISM_Init( status )
  
    ! --- in/out -----------------------------------------
    
    integer, intent(out)         ::  status
    
    ! --- const -----------------------------------------
    
    character(len=*), parameter  ::  rname = mname//', PRISM_Init'
    
    ! --- begin -----------------------------------------
    
    if ( echam_lm /= echam_lm_FIXED ) then
      write (*,'("ERROR - levels in echam/toy do not match with levels coded here ...")')
      write (*,'("ERROR -   echam/toy lm : ",i6)') echam_lm
      write (*,'("ERROR -   TM5       lm : ",i6)') echam_lm_FIXED
      write (*,'("ERROR in ",a)') rname; status=1; return
    end if
  
    if ( echam_T /= echam_T_FIXED ) then
      write (*,'("ERROR - spectral resolution in echam/toy do not match with levels coded here ...")')
      write (*,'("ERROR -   echam/toy T : ",i6)') echam_T
      write (*,'("ERROR -   TM5       T : ",i6)') echam_T_FIXED
      write (*,'("ERROR in ",a)') rname; status=1; return
    end if

    ! echam spectral fields: number of elements:  
    echam_shn = (echam_T+1)*(echam_T+2)/2
    
    ! NOTE: the following statements could not be used,
    ! since TM5 Init is called AFTER echo/toy filled the arrays ...
      
    !! allocate fields:
    !allocate( echam_lsp_tmin (2,echam_shn)          )
    !allocate( echam_lsp_tmid (2,echam_shn)          )
    !allocate( echam_lsp_tplus(2,echam_shn)          )
    !allocate( echam_vor_tmid (2,echam_shn,echam_lm) )
    !allocate( echam_div_tmid (2,echam_shn,echam_lm) )

    ! ok
    status = 0

  end subroutine PRISM_Init
    
    
  ! ***
  
  
  subroutine PRISM_Done( status )
  
    ! --- in/out -----------------------------------------
    
    integer, intent(out)         ::  status
    
    ! --- const -----------------------------------------
    
    character(len=*), parameter  ::  rname = mname//', PRISM_Done'
    
    ! --- begin -----------------------------------------
  
!    ! deallocate fields:
!    deallocate( echam_lsp_tmin  )
!    deallocate( echam_lsp_tmid  )
!    deallocate( echam_lsp_tplus )
!    deallocate( echam_vor_tmid  )
!    deallocate( echam_div_tmid  )

    ! ok
    status = 0

  end subroutine PRISM_Done


  ! ***
  
  
!  subroutine PRISM_GetInVarID( fieldname, id, status )
!  
!    ! --- in/out -----------------------------------------
!    
!    character(len=*), intent(in)   ::  fieldname
!    integer, intent(out)           ::  id
!    integer, intent(out)           ::  status
!    
!    ! --- const -----------------------------------------
!    
!    character(len=*), parameter  ::  rname = mname//', PRISM_GetInVarID'
!    
!    ! --- begin -----------------------------------------
!    
!    ! select var_id for this input field:
!    select case ( fieldname )
!      case ( 'sst' ); id = 1
!      case default
!        write (*,'("ERROR - unsupported field name `",a,"`")') fieldname
!        write (*,'("ERROR in ",a)') rname; status=1; return
!    end select
!    
!    ! ok
!    status = 0
!  
!  end subroutine PRISM_GetInVarID
!
!
!  ! ***
!  
!  
!  subroutine PRISM_GetInVar( var_id, status, ggN )
!  
!    ! --- in/out -----------------------------------------
!    
!    integer, intent(in)              ::  var_id
!    integer, intent(out)             ::  status
!    
!    integer, intent(out), optional   ::  ggN
!    
!    ! --- const -----------------------------------------
!    
!    character(len=*), parameter  ::  rname = mname//', PRISM_GetInVar'
!    
!    ! --- begin -----------------------------------------
!    
!    ! return gg size ?
!    if ( present(ggN) ) then
!      write (*,'("ERROR - ggN not implemented yet ...")')
!      write (*,'("ERROR in ",a)') rname; status=1; return
!      ggN = 0
!    end if
!    
!    ! ok
!    status = 0
!  
!  end subroutine PRISM_GetInVar
!
!
!  ! ***
!  
!
!  subroutine PRISM_GetField( fieldname, itap_sec, gg, status )
!  
!    ! --- in/out -----------------------------------------
!    
!    character(len=*), intent(in)   ::  fieldname
!    integer, intent(in)            ::  itap_sec
!    real, intent(out)              ::  gg(:)
!    integer, intent(out)           ::  status
!    
!    ! --- const -----------------------------------------
!    
!    character(len=*), parameter  ::  rname = mname//', PRISM_GetField'
!
!    ! --- begin -----------------------------------------
!    
!!    ! select var_id for this input field:
!!    select case ( fieldname )
!!
!!      case ( 'sst' )
!!      
!!        ! receive field ...
!!        call prism_get_proto (il_var_id_in(var_id), itap_sec, sst, status)
!!        write (il_mparout,fmt=*) 'itap_sec, statussst=',itap_sec, status
!!        if ( (status /= PRISM_ok) .and. (status < PRISM_recvd) ) then
!!          write (il_mparout,'("Pb in reading ",A8)') cl_read(var_id)
!!          write (il_mparout,'("Time is ",I8)') itap_sec
!!          write (il_mparout,'("Error code is ",I2)') status
!!          call prism_abort_proto(il_comp_id, 'tm5.f90','abort3') 
!!        end if
!!
!!        ! type conversion ...
!!        gg = sst
!!
!!      case default
!!
!!        write (il_mparout,'("ERROR - unsupported field name `",a,"`")') fieldname
!!        call prism_abort_proto(il_comp_id, 'tm5.f90','abort3') 
!!
!!    end select
!    gg = 0.0
!  
!    ! ok
!    status = 0
!  
!  end subroutine PRISM_GetField



  ! ==============================================================
  

  subroutine mf_Init( mf, paramkey, status )
  
    ! --- in/out ----------------------------
    
    type(TMeteoFile_prism), intent(out)   ::  mf
    character(len=*), intent(in)          ::  paramkey
    integer, intent(out)                  ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  name = mname//', mf_Init'
    
    ! --- begin --------------------------------
    
    ! dummy contents ...
    mf%id = 0
    
!    ! select var id in coupler for requested field:
!    call PRISM_GetInVarID( paramkey, mf%id, status )
!    if (status/=0) then; write (*,'("ERROR in ",a)') name; status=1; return; end if

    ! ok
    status = 0
 
  end subroutine mf_Init
  
  
  ! ***
  
  
  subroutine mf_Done( mf, status )
  
    ! --- in/out ------------------------------------

    type(TMeteoFile_prism), intent(inout) ::  mf
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  name = mname//', mf_Done'
    
    ! --- begin -------------------------------------
    
    ! ok
    status = 0
    
  end subroutine mf_Done
  


  ! ***
  
  
  subroutine mf_ReadRecord( mf, paramkey, t1, t2, nuv, gridtype, levi, &
                                shi, sh, lnsp_sh, &
                                status )

                                !lli, ll, sp_ll, &
                                !ggi, gg, sp_gg, &
    
    use PArray, only : pa_SetShape
    use GO, only : TDate, PrintDate
    use GO, only : operator(+), operator(-), operator(/), operator(==)
    use Grid, only : TllGridInfo, TggGridInfo, TshGridInfo, TLevelInfo
    use Grid, only : Init, Set

    ! --- in/out -------------------------------
    
    type(TMeteoFile_prism), intent(inout)  ::  mf
    character(len=*), intent(in)         ::  paramkey
    type(TDate), intent(in)              ::  t1, t2
    character(len=1), intent(in)         ::  nuv

    character(len=2), intent(out)        ::  gridtype
    type(TLevelInfo), intent(out)        ::  levi

    !type(TllGridInfo), intent(inout)     ::  lli
    !real, pointer                        ::  ll(:,:,:)
    !real, pointer                        ::  sp_ll(:,:)

    !type(TggGridInfo), intent(inout)     ::  ggi
    !real, pointer                        ::  gg(:,:)
    !real, pointer                        ::  sp_gg(:)

    type(TshGridInfo), intent(inout)     ::  shi
    complex, pointer                     ::  sh(:,:)
    complex, pointer                     ::  lnsp_sh(:)

    integer, intent(out)                 ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  name = mname//', mf_ReadRecord'
    
    ! --- local -----------------------------------
    
    type(TDate)    ::  tmid
    
    ! --- begin ---------------------------------

    ! mid time of [t1,t2]
    tmid = t1 + (t2-t1)/2

    !
    ! return spectral fields:
    !
    
    select case ( paramkey )

      case ( 'LNSP' )
      
        ! not for time intervals ...
        if ( .not. (t1 == t2) ) then
          write (*,'("ERROR - LNSP not valid in intervals; request with t1=t2 ...")')
          call PrintDate( 'ERROR -   t1 : ', t1 )
          call PrintDate( 'ERROR -   t2 : ', t2 )
          write (*,'("ERROR in ",a)') name; status=1; return
        end if
      
        ! return spectral field:
        gridtype = 'sh'

        ! intialize spherical harmonic field info:
        call Init( shi )
        call Set( shi, echam_T )
        
        ! fill log surface pressure:
        call pa_SetShape( lnsp_sh, echam_shn )
        if ( t1 == echam_tmin ) then
          lnsp_sh = cmplx(echam_lsp_tmin(1,:),echam_lsp_tmin(2,:))
        else if ( t1 == echam_tmid ) then
          lnsp_sh = cmplx(echam_lsp_tmid(1,:),echam_lsp_tmid(2,:))
        else if ( t1 == echam_tplus ) then
          lnsp_sh = cmplx(echam_lsp_tplus(1,:),echam_lsp_tplus(2,:))
        else
          write (*,'("ERROR - requested times do not match with stored echam fields:")')
          call PrintDate( 'ERROR -   t1, t2      : ', t1 )
          call PrintDate( 'ERROR -   echam tmin  : ', echam_tmin  )
          call PrintDate( 'ERROR -   echam tmid  : ', echam_tmid  )
          call PrintDate( 'ERROR -   echam tplus : ', echam_tplus )
          write (*,'("ERROR in ",a)') name; status=1; return
        end if
      
      case ( 'VO', 'D', 'T' )
      
        ! correct time ?
        if ( .not. (tmid == echam_tmid) ) then
          write (*,'("ERROR - requested time interval does not match with stored echam fields:")')
          call PrintDate( 'ERROR -   t1         : ', t1 )
          call PrintDate( 'ERROR -   t2         : ', t2 )
          call PrintDate( 'ERROR -   echam tmid : ', echam_tmid )
          write (*,'("ERROR in ",a)') name; status=1; return
        end if
      
        ! return spectral field:
        gridtype = 'sh'

        ! intialize spherical harmonic field info:
        call Init( shi )
        call Set( shi, echam_T )
        
        ! fill log surface pressure:
        call pa_SetShape( lnsp_sh, echam_shn )
        lnsp_sh = cmplx(echam_lsp_tmid(1,:),echam_lsp_tmid(2,:))
        
        ! fill 3D array
        call pa_SetShape( sh, echam_shn, echam_lm )
        select case ( paramkey )
          case ( 'VO' )
            sh = cmplx(echam_vor_tmid(1,:,:),echam_vor_tmid(2,:,:))
          case ( 'D' )
            sh = cmplx(echam_div_tmid(1,:,:),echam_div_tmid(2,:,:))
          case ( 'T' )
            sh = cmplx(echam_temper_tmid(1,:,:),echam_temper_tmid(2,:,:))
          case default
            write (*,'("ERROR - unsupported sh paramkey `",a,"`")') trim(paramkey)
            write (*,'("ERROR in ",a)') name; status=1; return
        end select

      case default

        write (*,'("ERROR - unsupported paramkey `",a,"`")') trim(paramkey)
        write (*,'("ERROR in ",a)') name; status=1; return

    end select
    
    !
    ! return level info
    !
    
    select case ( echam_lm )
      case ( 19 )
        call Init( levi, echam_lm, echam_lev19_a, echam_lev19_b, status )
        if ( status/= 0 ) then; write (*,'("ERROR in ",a)') name; status=1; return; end if
      case ( 39 )
        call Init( levi, echam_lm, echam_lev39_a, echam_lev39_b, status )
        if ( status/= 0 ) then; write (*,'("ERROR in ",a)') name; status=1; return; end if
      case default
        write (*,'("ERROR - unsupported echam lm : ",i4)') echam_lm
        write (*,'("ERROR in ",a)') name; status=1; return
    end select
    
    ! ok
    status = 0

  end subroutine mf_ReadRecord
  

  
end module tmm_mf_prism
